@@ -10,7 +10,8 @@
 #
 #################################################################
 
-read.px <- function( filename , encod.from="ISO_8859-1", encod.to="UTF-8") {
+read.px <- function( filename , encod.from = "ISO_8859-1", encod.to = "UTF-8",
+                    na.strings=c('"."', '".."', '"..."', '"...."')) {
 
     ## auxiliary functions ##
 
@@ -19,14 +20,14 @@ read.px <- function( filename , encod.from="ISO_8859-1", encod.to="UTF-8") {
     }
 
     clean.spaces <- function( x ){
-        x <- gsub("^[[:space:]]+","",x) # elimina espacios en blanco por delante
-        x <- gsub("[[:space:]]+$","",x) # elimina espacios en blanco por detras
+        x <- gsub( "^[[:space:]]+", "", x ) # elimina espacios en blanco por delante
+        x <- gsub( "[[:space:]]+$", "", x ) # elimina espacios en blanco por detras
         x
     }
 
     get.attributes <- function( x ){
         x <- gsub( "([A-Z-]*)\\((.*)\\).*", "\\1;\\2", x ) ## parte etiqueta y atributo con ";"
-        x <- strsplit( x, ";" )
+        x <- strsplit( x, ";")
         x <- lapply( x, function( x ) c( x, rep( "value", 2 - length( x ) ) ) )
         x <- do.call( rbind, x )
         x[,2] <- unquote( x[,2] )
@@ -36,30 +37,36 @@ read.px <- function( filename , encod.from="ISO_8859-1", encod.to="UTF-8") {
     break.clean <- function( x, sep = '\\"' ) {
         #x <- strsplit( unquote( x ), sep )[[1]]
 
-        x <- strsplit( x, sep )[[1]]
-        x <- clean.spaces( x ) 
+        x <- strsplit( x, sep, useBytes=TRUE)[[1]]
+        if (sep != " ") x <- clean.spaces( x )
         x <- x[ x != "" ]
         x <- x[ x != "," ]
         x
     }
 
 
-    cleanDat <- function(x){
+    ## cleanDat <- function( x ){
+    ##     x <- break.clean( x, " " )
+    ##     x <- gsub( ".*,", "", x )           # eliminates keys part in files with KEYS argument
+    ##     x <- gsub( "\"\\.\\.\"","NA", x )
+    ##     x <- gsub( "\"\\.\"" ,"NA", x ) 
+    ##     as.numeric( x[ x != "" ] )
+    ## }
+
+    get.keys <- function( x, codes, values, keys ){
         x <- break.clean( x, " " )
-        x <- gsub( "\"..\"","NA", x )     # cambia los valores para
-                                          # mising ".." ## creo que es
-                                          # incorrecto: deber?a ser la
-        x <- gsub( "\".\"" ,"NA", x )     # cambia los valores para
-                                          # mising "." ## expresi?n
-                                          # regular "\\.\\." para el
-                                          # doble punto.
-                                            ## adem?s, hay que tener
-                                            ## en cuenta que puede
-                                            ## sustituir el punto
-                                            ## decimal!  igual habr?a
-                                            ## que sustituir " \\.+ "
-                                            ## por NA
-        as.numeric( x[ x != "" ] )
+        x <- x[ grep( ",", x ) ]                # keep just components having a comma 
+        x <- unquote( x )
+        x <- do.call( rbind, sapply( x, break.clean, ",", simplify = F ) )
+        x <- data.frame( x[, -ncol( x ) ] )               # drops the data part
+        rownames( x ) <- NULL
+        colnames( x ) <- rev( names( keys ) )
+        
+        encoded.vars <- names( keys )[ keys == "CODES" ]
+        for( var.name in encoded.vars ){
+            x[[var.name]] <- values[[ var.name ]] [ match( x[[ var.name ]], codes[[ var.name ]] ) ]
+        }
+        x
     }
 
     make.list <- function( dat, my.label ){
@@ -79,8 +86,8 @@ read.px <- function( filename , encod.from="ISO_8859-1", encod.to="UTF-8") {
                                              ## platforms; maybe a
                                              ## different encoding is
                                              ## required
-    a <- unlist( strsplit( a, ";") )
-    a <- do.call( rbind, strsplit( a, "=" ) )
+    a <- unlist( strsplit( a, ";", fixed=TRUE, useBytes = TRUE ) )
+    a <- do.call( rbind, strsplit( a, "=", fixed=TRUE, useBytes=TRUE) )
 
     a <- data.frame( cbind( get.attributes( a[, 1]), a[, 2] ) )
     colnames( a ) <- c( "label", "attribute", "value" )
@@ -93,15 +100,24 @@ read.px <- function( filename , encod.from="ISO_8859-1", encod.to="UTF-8") {
 
     px <- sapply( unique( a$label ), function( label ) make.list( a, label ), simplify = FALSE )
 
-
     # turns data values into an R vector
-    px$DATA$value    <- cleanDat( px$DATA$value )
     px$STUB$value    <- make.names( break.clean( px$STUB$value ) )
     px$HEADING$value <- make.names( break.clean( px$HEADING$value ) )
 
     px$VALUES <- lapply( px$VALUES, break.clean )
     px$CODES  <- lapply( px$CODES, break.clean )
 
+    # only if data contains the KEYS keyword
+    # such datasets have a different data format and require an extra field to
+    # keep track of which rows are present in the data (see format document for details)
+    #if( "KEYS" %in% names( px ) )
+    #    px$internal.keys <- get.keys( px$DATA$value, px$CODES, px$VALUES, px$KEYS )
+
+##    px$DATA$value    <- cleanDat( px$DATA$value )
+    dat <- textConnection(px$DATA$value) #much faster than with cleanDat (strsplit)
+    px$DATA$value <- scan(dat, na.strings=na.strings, quiet=TRUE)
+    close(dat)
+    
     class( px ) <- "px"
     return( px )
 }
